from typing import Optional

import podgenai.exceptions


def get_confirmation(next_task: Optional[str] = None) -> None:
    """Receive input confirmation from the user, optionally for the specified next task.

    If confirmation is refused, `InputError` is raised.
    """
    task_prompt = f" with {next_task}" if next_task else ""
    user_prompt = f"Continue{task_prompt}? [y/n]: "
    while True:
        response = input(user_prompt)
        response = response.strip().lower()
        match response:
            case "y" | "yes":
                break
            case "n" | "no":
                raise podgenai.exceptions.InputError("User canceled.")



import os
from pathlib import Path
from typing import Optional

import dotenv
import openai
import pathvalidate

import podgenai.exceptions
from podgenai.config import PROMPTS
from podgenai.util.binascii import hasher
from podgenai.util.sys import print_warning

dotenv.load_dotenv()

ChatCompletion = openai.types.chat.chat_completion.ChatCompletion
OpenAI = openai.OpenAI

MAX_TTS_INPUT_LEN = 4096
MODELS = {
    "text": "gpt-4-0125-preview",
    # Notes:
    #   As of 2024-05, gpt-4o (gpt-4o-2024-05-13) is not used because it was observed to hallucinate significantly, whereas gpt-4-turbo-preview (gpt-4-0125-preview) doesn't.
    #   As of 2024-04, gpt-4-turbo (gpt-4-turbo-2024-04-09) is not used because it was observed to produce slightly lesser content than gpt-4-turbo-preview (gpt-4-0125-preview).
    #   gpt-4 is not used because it is much older in its training data.
    "tts": "tts-1",  # Note: tts-1-hd is twice as expensive, and has a more limited concurrent usage quota resulting in openai.RateLimitError, thereby making it undesirable.
}
TTS_VOICE_MAP = {
    "default": "alloy",
    "neutral": "alloy",  # Unsolicited but observed.
    "female": "nova",
    "male": "onyx",
    "sensitive": "echo",  # Supported for testing.
    "emotive": "echo",
}


def ensure_openai_key() -> None:
    """Raise `EnvError` if the environment variable OPENAI_API_KEY is unavailable."""
    if not os.environ.get("OPENAI_API_KEY"):
        raise podgenai.exceptions.EnvError("The environment variable OPENAI_API_KEY is unavailable. It can optionally be defined in an .env file.")


def get_openai_client() -> OpenAI:
    """Return the OpenAI client."""
    return OpenAI()


def get_completion(prompt: str, *, client: Optional[OpenAI] = None) -> ChatCompletion:
    """Return the completion for the given prompt."""
    if not client:
        client = get_openai_client()
    # print(f"Requesting completion for prompt of length {len(prompt)}.")
    completion = client.chat.completions.create(model=MODELS["text"], messages=[{"role": "user", "content": prompt}])
    # Note: Specifying max_tokens=4096 with gpt-4-turbo-preview did not benefit in increasing output length, and a higher value is disallowed. Ref: https://platform.openai.com/docs/api-reference/chat/create
    return completion


def get_multipart_messages(prompt: str, *, max_completions: int = 10, client: Optional[OpenAI] = None, update_prompt: bool = False, continuation: str = PROMPTS["continuation_next"]) -> list[dict]:
    """Return the multipart completion messages for the given initial prompt.

    After the initial completion, continuation prompts are subsequently given, either until the assistant is done, or until a maximum of `max_completions` are received, whichever is first.

    If `update_prompt` is True, the initial prompt is appended a default note that continuation prompts will subsequently be sent.
    If `update_prompt` is False, the initial prompt can be provided with an included continuation note.

    If `continuation` is specified, it is used iteratively as the continuation prompt, otherwise a default continuation prompt is used.
    """
    if update_prompt:
        prompt = prompt + "\n\n" + PROMPTS["continuation_first"]
    endings = ("Done", "Done.", "done", "done.")

    if not client:
        client = get_openai_client()

    messages = [{"role": "user", "content": prompt}]
    for completion_num in range(1, max_completions + 1):
        # print(f"Requesting completion {completion_num} for initial prompt of length {len(prompt)}.")
        completion = client.chat.completions.create(model=MODELS["text"], messages=messages)
        content = get_content(prompt="", completion=completion)
        messages.append({"role": "assistant", "content": content})

        if content in endings:
            print(f"Completion {completion_num} is an ending.")
            return messages
        else:
            for ending in endings:
                if content.endswith((f" {ending}", f"\n{ending}")):
                    print(f"Completion {completion_num} has an ending.")
                    return messages

        if completion_num == max_completions:
            print_warning(f"The quota of a maximum of {max_completions} completions is exhausted for initial prompt of length {len(prompt)}.")
            return messages

        messages.append({"role": "user", "content": continuation})

    assert False


def get_content(prompt: str, *, client: Optional[OpenAI] = None, completion: Optional[ChatCompletion] = None) -> str:
    """Return the content for the given prompt."""
    if not completion:
        completion = get_completion(prompt, client=client)
    content = completion.choices[0].message.content
    content = content.strip()
    assert content
    return content


def get_multipart_content(prompt: str, **kwargs) -> str:
    """Return the multipart completion content for the given initial prompt.

    Additional keyword arguments are forwarded to `get_multipart_messages`.

    The completions are joined using paragraph breaks (double line breaks).
    """
    endings = ("Done", "Done.", "done", "done.")
    messages = get_multipart_messages(prompt, **kwargs)

    completions = []
    for message_count, message in enumerate(messages, start=1):
        if message["role"] != "assistant":
            continue
        completion = message["content"]
        assert completion == completion.strip()
        if completion in endings:
            assert message_count == len(messages), {"prompt": prompt, "messages": messages, "message_count": message_count, "completion": completion}
            break
        for ending in endings:
            if completion.endswith((f" {ending}", f"\n{ending}")):
                completion = completion.removesuffix(ending).rstrip()
                break
        completions.append(completion)

    return "\n\n".join(completions).strip()


def get_cached_content(prompt: str, *, strategy: str = "oneshot", read_cache: bool = True, cache_key_prefix: str, cache_path: Path, **kwargs) -> str:
    """Return the content for the given prompt using the disk cache if available, otherwise normally.

    Params:
    * `strategy`:
        If `strategy` is 'oneshot', the assistant is requested only one output, which is usually sufficient.
        If `strategy` is 'multishot', the assistant is permitted multiple outputs up to a limit.
    * `read_cache`: If `True`, the disk cache is read if available. If `False`, the disk cache is not read, and it will be written or overwritten.
    * `cache_key_prefix`: Friendly identifying name of request, used in filename in cache directory. Deduplication by prompt is done by this function; it does not have to be done externally.
    * `cache_path`: Cache directory.

    Additional keyword arguments are forwarded to the respective underlying function, namely `get_content` for 'oneshot', and `get_multipart_content` for 'multishot'.
    """
    cache_key_prefix = cache_key_prefix.strip()
    assert cache_key_prefix
    assert cache_path.is_dir()

    sanitized_cache_key_prefix = pathvalidate.sanitize_filename(cache_key_prefix, platform="auto")
    assert sanitized_cache_key_prefix
    cache_key = f"{sanitized_cache_key_prefix} ({strategy}) [{hasher(prompt)}].txt"
    cache_file_path = cache_path / cache_key
    pathvalidate.validate_filepath(cache_file_path, platform="auto")

    if read_cache and cache_file_path.exists():
        assert cache_file_path.is_file()
        content = cache_file_path.read_text().rstrip()  # rstrip is used in case the file is manually modified in an editor which adds a trailing newline.
        print(f"Read completion from disk for: {cache_key_prefix}")
    else:
        content_getter = {"oneshot": get_content, "multishot": get_multipart_content}[strategy]
        print(f"Requesting completion for: {cache_key_prefix}")
        content = content_getter(prompt, **kwargs)
        print(f"Received completion for: {cache_key_prefix}")
        cache_file_path.write_text(content)

    assert content == content.rstrip()
    return content


def write_speech_audio(text: str, path: Path, *, voice: str = "default", client: Optional[OpenAI] = None) -> None:
    """Write the speech audio file for the given prompt to the given file path.

    The prompt must not be longer than 4096 characters, as this is the maximum supported length by the client.

    `voice` can be one of the keys or values in TTS_VOICE_MAP, or one of the other supported voices.
    """
    assert path.suffix == ".mp3"
    if not client:
        client = get_openai_client()

    mapped_voice = TTS_VOICE_MAP.get(voice, voice)
    voice_str = voice if (voice == mapped_voice) else f"{voice} ({mapped_voice})"

    print(f"Requesting speech audio in {voice_str} voice for: {path.stem}")
    response = client.audio.speech.create(model=MODELS["tts"], voice=mapped_voice, input=text)
    # relative_path = path.relative_to(Path.cwd())
    # print(f"Writing speech to: {relative_path}")
    response.stream_to_file(path)
    assert path.exists(), path
    # print(f"Wrote speech to: {relative_path}")
    print(f"Received speech audio in {voice_str} voice for: {path.stem}")


def ensure_speech_audio(text: str, path: Path, **kwargs) -> None:
    """Ensure the speech audio file for the given text to the given file path.

    Additional keyword arguments are forwarded to `write_speech`.
    """

    if path.exists():
        assert path.is_file()
        print(f"Speech audio file exists on disk for: {path.stem}")
        return
    write_speech_audio(text, path=path, **kwargs)



import concurrent.futures
import contextlib
import io
from typing import Optional

import podgenai.exceptions
from podgenai.config import MAX_CONCURRENT_WORKERS, PROMPTS
from podgenai.util.openai import get_cached_content
from podgenai.work import get_topic_work_path
from podgenai.util.sys import print_error, print_warning


def is_subtopics_list_valid(subtopics: list[str]) -> bool:
    """Return true if the subtopics are structurally valid, otherwise false.

    A validation error is printed if a subtopic is invalid.
    """
    if not subtopics:
        print_error("No subtopics exist.")
        return False

    seen = set()
    for num, subtopic in enumerate(subtopics, start=1):
        if subtopic != subtopic.strip():
            print_error(f"Subtopic {num} is invalid because it has leading or trailing whitespace: {subtopic!r}")
            return False

        expected_num_prefix = f"{num}. "
        if not subtopic.startswith(expected_num_prefix):
            print_error(f"Subtopic {num} is invalid because it is not numbered correctly: {subtopic}")
            return False

        subtopic_name = subtopic.removeprefix(expected_num_prefix).strip()
        if not subtopic_name:
            print_error(f"Subtopic {num} is invalid because it has no value: {subtopic}")
            return False

        if subtopic_name != subtopic_name.lstrip():
            print_error(f"Subtopic {num} is invalid because its name has leading whitespace: {subtopic!r}")
            return False

        if subtopic_name in seen:
            print_error(f"Subtopic {num} is invalid because its name is a duplicate: {subtopic}")
            return False
        seen.add(subtopic_name)

    return True


def list_subtopics(topic: str, max_attempts: int = 2) -> list[str]:
    """Return the list of subtopics for the given topic.

    Params:
    * `max_attempts`: If greater than 1, and if the first attempt obtains no subtopics, subsequent attempt(s) will be made. Only the first attempt tries to read from the disk cache.

    `LanguageModelOutputError` is raised if the model output is structurally invalid.
    """
    prompt_name = "list_subtopics"
    prompt = PROMPTS[prompt_name].format(topic=topic)
    none_subtopics = ("none", "none.")
    invalid_subtopics = ("", *none_subtopics)

    for num_attempt in range(1, max_attempts + 1):
        subtopics = get_cached_content(prompt, read_cache=num_attempt == 1, cache_key_prefix=f"0. {prompt_name}", cache_path=get_topic_work_path(topic))
        assert subtopics, subtopics

        if subtopics.lower() in none_subtopics:
            if num_attempt == max_attempts:
                raise podgenai.exceptions.LanguageModelOutputError(f"No subtopics exist after {max_attempts} attempts for topic: {topic}")
            else:
                continue

        assert subtopics.lower() not in none_subtopics, subtopics
        subtopics = [s.strip() for s in subtopics.splitlines() if s.strip().lower() not in invalid_subtopics]  # Note: A terminal "None" line has been observed with valid subtopics before it.

        error = io.StringIO()
        with contextlib.redirect_stderr(error):
            subtopics_list_is_valid = is_subtopics_list_valid(subtopics)
        if not subtopics_list_is_valid:
            error = error.getvalue().rstrip().removeprefix("Error: ")
            if num_attempt == max_attempts:
                raise podgenai.exceptions.LanguageModelOutputError(error)
            else:
                print_warning(f"Fault in attempt {num_attempt} of {max_attempts}: {error}")
                # Note: This condition has been observed with the subtopic list not being numbered correctly.
                continue

        break

    assert subtopics
    assert is_subtopics_list_valid(subtopics)
    return subtopics


def get_subtopic(*, topic: str, subtopics: list[str], subtopic: str, strategy: str = "oneshot") -> str:
    """Return the full text for a given subtopic within the context of the given topic and list of subtopics."""
    assert subtopic[0].isdigit()  # Is numbered.
    common_kwargs = {"strategy": strategy, "cache_key_prefix": subtopic, "cache_path": get_topic_work_path(topic)}
    match strategy:
        case "oneshot":
            prompt = PROMPTS["generate_subtopic"].format(optional_continuation="", topic=topic, subtopics="\n".join(subtopics), numbered_subtopic=subtopic)
            subtopic = get_cached_content(prompt, **common_kwargs)
        case "multishot":  # Observed to never really benefit or produce longer content relative to oneshot.
            prompt = PROMPTS["generate_subtopic"].format(optional_continuation="\n\n" + PROMPTS["continuation_first"], topic=topic, subtopics="\n".join(subtopics), numbered_subtopic=subtopic)
            subtopic = get_cached_content(prompt, **common_kwargs, max_completions=5, update_prompt=False)
        case _:
            assert False, strategy
    return subtopic.rstrip()


def get_subtopics_texts(*, topic: str, subtopics: Optional[list[str]] = None) -> dict[str, str]:
    """Return the ordered full text for all subtopics within the context of the given topic and optional ordered list of subtopics.

    If the list of subtopics is not provided, it is read.
    """
    if not subtopics:
        subtopics = list_subtopics(topic)
    if MAX_CONCURRENT_WORKERS == 1:
        subtopics_texts = {s: get_subtopic(topic=topic, subtopics=subtopics, subtopic=s) for s in subtopics}
    else:
        assert MAX_CONCURRENT_WORKERS > 1
        with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_CONCURRENT_WORKERS) as executor:
            fn_get_subtopic = lambda subtopic: get_subtopic(topic=topic, subtopics=subtopics, subtopic=subtopic)
            subtopics_texts = {s: text for s, text in zip(subtopics, executor.map(fn_get_subtopic, subtopics))}
    return subtopics_texts


def get_subtopics_speech_texts(*, topic: str, subtopics: Optional[list[str]] = None) -> dict[str, str]:
    """Return the ordered speech text for all subtopics within the context of the given topic and optional ordered list of subtopics.

    If the list of subtopics is not provided, it is read.
    """
    if not subtopics:
        subtopics = list_subtopics(topic)
    subtopics_texts = get_subtopics_texts(topic=topic, subtopics=subtopics)

    subtopics_speech_texts = {subtopic_name: f'Section {subtopic_name.replace('.', ':', 1)}:\n\n{subtopic_text} {{pause}}' for subtopic_name, subtopic_text in subtopics_texts.items()}
    # Note: A pause at the beginning is skipped by the TTS generator, but it is not skipped if at the end, and so it is kept at the end.

    subtopics_speech_texts[subtopics[0]] = f'{topic}:\n\n{{pause}}\n{PROMPTS['tts_disclaimer']} {{pause}}\n\n{subtopics_speech_texts[subtopics[0]]}'
    # Note: TTS disclaimer about AI generated audio is required by OpenAI as per https://platform.openai.com/docs/guides/text-to-speech/do-i-own-the-outputted-audio-files
    # Note: It has proven more reliable for the pause to be structured in this way for section 1, rather than be in the leading topic line.

    subtopics_speech_texts[subtopics[-1]] = f"{subtopics_speech_texts[subtopics[-1]]}\n\nThe end."
    # Note: "The end." sounds better with the paragraph break before it.

    return subtopics_speech_texts



import concurrent.futures
from pathlib import Path

import pathvalidate

from podgenai.config import MAX_CONCURRENT_WORKERS
from podgenai.util.binascii import hasher
from podgenai.util.openai import ensure_speech_audio, MAX_TTS_INPUT_LEN
from podgenai.util.semantic_text_splitter import semantic_split
from podgenai.work import get_topic_work_path


def get_speech_tasks(subtopics_speech_texts: dict[str, str], *, topic: str, voice: str) -> dict[Path, str]:
    """Return the pairs of texts to write to file paths as audio."""
    work_path = get_topic_work_path(topic)
    tts_tasks = {}
    for part_idx, (part_title, part) in enumerate(subtopics_speech_texts.items()):
        part_dedup_hash = hasher(part)
        part_stem = f"{part_title} ({voice}) [{part_dedup_hash}]"
        part_stem = pathvalidate.sanitize_filename(part_stem, platform="auto")
        if len(part) <= MAX_TTS_INPUT_LEN:
            part_path = work_path / f"{part_stem}.mp3"
            pathvalidate.validate_filepath(part_path, platform="auto")
            tts_tasks[part_path] = part
        else:
            portions = semantic_split(part, MAX_TTS_INPUT_LEN)
            for portion_num, portion in enumerate(portions, start=1):
                assert len(portion) <= MAX_TTS_INPUT_LEN
                portion_path = work_path / f"{part_stem} ({portion_num}).mp3"
                pathvalidate.validate_filepath(portion_path, platform="auto")
                tts_tasks[portion_path] = portion
    return tts_tasks


def ensure_speech_audio_files(parts: dict[Path, str], voice: str) -> None:
    """Ensure the speech audio files for the given part file path and text pairs.

    If a given file path already exists, it is not rewritten. If it does not exist, it is written.
    """
    if MAX_CONCURRENT_WORKERS == 1:
        for part_path, part_text in parts.items():
            ensure_speech_audio(part_text, path=part_path, voice=voice)
    else:
        assert MAX_CONCURRENT_WORKERS > 1
        with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_CONCURRENT_WORKERS) as executor:
            fn_ensure_speech_audio = lambda part_path: ensure_speech_audio(parts[part_path], path=part_path, voice=voice)
            list(executor.map(fn_ensure_speech_audio, parts))



from pathlib import Path

import pathvalidate

from podgenai.config import WORK_PATH
from podgenai.content.topic import ensure_topic_is_valid


def get_topic_work_path(topic: str, create: bool = True) -> Path:
    """Return the working directory for the given topic.

    If `create` is True, the directory is created if it does not already exist.
    """
    ensure_topic_is_valid(topic)
    work_path = WORK_PATH / pathvalidate.sanitize_filename(topic, platform="auto")
    pathvalidate.validate_filepath(work_path, platform="auto")
    if create:
        work_path.mkdir(parents=True, exist_ok=True)
    return work_path



class Error(Exception):
    """Base error."""


class EnvError(Error, EnvironmentError):
    """Environment variable availability error."""


class InputError(Error, ValueError):
    """Input validity error."""


class ModelOutputError(Error):
    """Model output error."""


class LanguageModelOutputError(ModelOutputError):
    """Language model output error."""



from podgenai.config import PROMPTS
from podgenai.util.openai import get_cached_content, TTS_VOICE_MAP
from podgenai.work import get_topic_work_path


def get_voice(topic: str, max_attempts: int = 3) -> str:
    """Return the suggested voice for the given topic.

    Params:
    * `max_attempts`: If greater than 1, and if the first attempt obtains an invalid voice, subsequent attempt(s) will be made. Only the first attempt tries to read from the disk cache.
    """
    # Note: More than a single attempt is necessary because an invalid value such as "emale" has at times been observed in the first attempt.
    prompt_name = "select_voice"
    prompt = PROMPTS[prompt_name].format(topic=topic)

    for num_attempt in range(1, max_attempts + 1):
        raw_voice = get_cached_content(prompt, read_cache=num_attempt == 1, cache_key_prefix=f"0. {prompt_name}", cache_path=get_topic_work_path(topic))
        voice = raw_voice.rstrip(".").lower()
        if voice in TTS_VOICE_MAP:
            break

    assert voice in TTS_VOICE_MAP, {"topic": topic, "raw_voice": raw_voice, "voice": voice, "TTS_VOICE_MAP": TTS_VOICE_MAP}
    return voice



import sys


def print_error(error: str) -> None:
    """Print error message to stderr."""
    print(f"Error: {error}", file=sys.stderr)


def print_warning(warning: str) -> None:
    """Print warning message to stderr."""
    print(f"Warning: {warning}", file=sys.stderr)



import os
from pathlib import Path

import dotenv

dotenv.load_dotenv()

CWD: Path = Path.cwd()
PACKAGE_PATH: Path = Path(__file__).parent
REPO_PATH: Path = PACKAGE_PATH.parent.parent

GiB = 1024**3
# DISKCACHE_PATH = REPO_PATH / ".diskcache"
# DISKCACHE_SIZE_LIMIT = 1 * GiB
# DISKCACHE = diskcache.FanoutCache(directory=str(DISKCACHE_PATH), timeout=1, size_limit=DISKCACHE_SIZE_LIMIT)
MAX_CONCURRENT_WORKERS = int(os.environ.get("PODGENAI_OPENAI_MAX_WORKERS", 16))  # Note: Default value is documented in readme.
assert MAX_CONCURRENT_WORKERS >= 1
PROMPTS: dict[str, str] = {p.stem: p.read_text().strip() for p in (PACKAGE_PATH / "prompts").glob("*.txt")}
WORK_PATH: Path = CWD / "work"



from typing import Optional

import fire
import os
from pathlib import Path

import podgenai.exceptions
from podgenai.podgenai import generate_media
from podgenai.content.topic import get_topic, ensure_topic_is_valid
from podgenai.util.openai import ensure_openai_key
from podgenai.util.sys import print_error


def main(topic: Optional[str] = None, path: Optional[Path] = None, confirm: bool = False) -> None:
    """Generate and write an audiobook podcast mp3 file for the given topic to the given output file path.

    Params:
    * `topic` (-t): Topic. If not given, the user is prompted for it.
    * `path (-p)`: Output file or directory path.
        If an intended file path, it must have an ".mp3" suffix. If a directory, it must exist, and the file name is auto-determined.
        If not given, the output file is written to the repo directory with an auto-determined file name.
    * `confirm`: Confirm before full-text and speech generation.
        If true, a confirmation is interactively sought after generating and printing the list of subtopics, before generating the full-text, and also before generating the speech. Its default is false.

    A nonzero exitcode exists if there is an error.
    """
    try:
        ensure_openai_key()

        if not topic:
            topic = get_topic()
        ensure_topic_is_valid(topic)

        if path:
            path = Path(path)

        if not isinstance(confirm, bool):
            raise podgenai.exceptions.InputError("`confirm` (-c) argument has an invalid value. No value is to explicitly be specified for it since it is a boolean.")

        generate_media(topic, output_path=path, confirm=confirm)
    except podgenai.exceptions.Error as exc:
        print_error(str(exc))
        print_error(f"Failed to generate for topic: {topic}")
        exit(1)
    except KeyboardInterrupt:
        print()  # This separates "^C" from the subsequent error.
        print_error("Interrupted by user.")
        os._exit(-2)  # Plain `exit` is not used because it may not immediately terminate, with background threads potentially still running.


if __name__ == "__main__":
    fire.Fire(main)



import binascii


def crc32(text: str) -> str:
    """Return the CRC32 hash of the given string as a hexadecimal string."""
    data: bytes = text.encode()
    crc: int = binascii.crc32(data)
    crc: str = f"{crc:08x}"
    assert len(crc) == 8
    return crc


hasher = crc32



import datetime
from pathlib import Path
import subprocess
from typing import Optional

import pathvalidate

from podgenai.config import CWD
from podgenai.work import get_topic_work_path


def get_default_output_filename(topic: str) -> str:
    """Return the default output filename for the given topic."""
    now = datetime.datetime.now().isoformat(timespec="seconds")
    output_filename = f"{now} {topic}.mp3"
    output_filename = pathvalidate.sanitize_filename(output_filename, platform="auto")
    return output_filename


def get_output_file_path(output_path: Optional[Path], *, topic: str) -> Path:
    """Return the validated output file path for the given topic."""
    if output_path is None:
        output_filename = get_default_output_filename(topic)
        output_path = CWD / output_filename
    else:
        output_path = output_path.expanduser().resolve()
        if output_path.is_dir():
            assert output_path.exists()
            output_filename = get_default_output_filename(topic)
            output_path = output_path / output_filename
        else:
            assert output_path.suffix == ".mp3"
            output_path.parent.mkdir(parents=True, exist_ok=True)
    pathvalidate.validate_filepath(output_path, platform="auto")
    return output_path


def merge_speech_paths(paths: list[Path], *, topic: str, output_path: Path) -> None:
    """Merge the ordered list of preexisting audio file paths for the given topic to a single audio file having the given output file path."""
    work_path = get_topic_work_path(topic)
    ffmpeg_paths = [str(p).replace("'", "'\\''") for p in paths]
    ffmpeg_filelist_path = work_path / "ffmpeg.list"
    ffmpeg_filelist_path.write_text("\n".join(f"file '{p}'" for p in ffmpeg_paths))
    print(f"Merging {len(paths)} speech parts.")
    subprocess.run(["ffmpeg", "-y", "-f", "concat", "-safe", "0", "-i", str(ffmpeg_filelist_path), "-c", "copy", "-loglevel", "error", str(output_path)], check=True)
    assert output_path.exists()
    print(f"Merged {len(paths)} speech parts.")
